#ifndef GEAPP
#define GEAPP

#include <algorithm>
#include <map>
#include <fstream>

#include "ist/ist_i3d.h"
#include "ist/ist_gl.h"

#include "wx/wx.h"
#include "wx/glcanvas.h"

#include "ground.h"



typedef unsigned char u_char;

inline float clamp(float f) { return std::min(0.99f, std::max(0.01f, f)); }



class Ground : public BGround
{
public:


	Ground() {}
	~Ground() {}


	bool save(const std::string& fname) {
		std::ofstream f(fname.c_str(), std::ios::binary);
		if(!f)
			return false;
		if(fname.rfind(".mqo")!=std::string::npos) {
			saveAsMQO(f);
			return true;
		}
		else if(fname.rfind(".obj")!=std::string::npos) {
			saveAsOBJ(f);
			return true;
		}
		else if(fname.rfind(".png")!=std::string::npos) {
			f.close();
			saveAsBitmap(fname);
			return true;
		}
		int version=101;
		f.write((char*)&version, sizeof(int));
		f.write((char*)min_edge.v, sizeof(float)*2);
		f.write((char*)max_edge.v, sizeof(float)*2);
		f.write((char*)&grid_size, sizeof(float));
		for(size_t i=0; i<vertex.size(); ++i) {
			f.write((char*)&vertex[i].pos.y, sizeof(float));
			f.write((char*)vertex[i].tex.v, sizeof(float)*2);
			f.write((char*)&vertex[i].attr, sizeof(int));
		}
		return true;
	}

	void saveAsMQO(std::ofstream& f) {
		f << "Metasequoia Document\n";
		f << "Format Text Ver 1.0\n\n";
		f << "Object \"ground\" {\n";
		f << "\tvertex " << vertex.size() << " {\n";
		for(size_t i=0; i<vertex.size(); ++i) {
			f << "\t\t" << vertex[i].pos.x << " " << vertex[i].pos.y << " " << vertex[i].pos.z << "\n";
		}
		f << "\t}\n";
		f << "\tface " << index.size()/4 << " {\n";
		for(size_t i=0; i<index.size()/4; ++i) {
			f << "\t\t" << 4 << " V(";
			f << index[i*4+3] << " " << index[i*4+2] << " " << index[i*4+1] << " " << index[i*4+0] << ") ";
			f << "UV(";
			for(int j=0; j<4; ++j) {
				f << vertex[index[i*4+3-j]].tex.x << " " << vertex[index[i*4+3-j]].tex.y;
				if(j!=3)
					f << " ";
			}
			f << ")\n";
		}
		f << "\t}\n";
		f << "}\n";
		f << "EoF\n";
	}

	void saveAsOBJ(std::ofstream& f) {
		for(size_t i=0; i<vertex.size(); ++i)
			f << "v " << vertex[i].pos.x << " " << vertex[i].pos.y << " " << vertex[i].pos.z << "\n";
		for(size_t i=0; i<vertex.size(); ++i)
			f << "vt " << vertex[i].tex.x << " " << vertex[i].tex.y << "\n";
		for(size_t i=0; i<index.size()/4; ++i) {
			f << "f ";
			for(int j=0; j<4; ++j) {
				f << index[i*4+j]+1 << "/" << index[i*4+j]+1;
				if(j!=3)
					f << " ";
			}
			f << "\n";
		}
	}

	void saveAsBitmap(const std::string& fname) {
		float low=0, high=0;
		getHeightRange(low, high);

		ist::Bitmap bmp(x_div, z_div);
		for(int i=0; i<z_div; ++i) {
			for(int j=0; j<x_div; ++j) {
				u_char p = u_char( ((*this)[i][j].pos.y-low)/(high-low)*255);
				bmp[i][j].r=bmp[i][j].g=bmp[i][j].b=p;
				bmp[i][j].a=255;
			}
		}
		bmp.save(fname);
	}

	void genBitmap(ist::Bitmap &bmp) {
		float low=0, high=0;
		getHeightRange(low, high);
		if(high-low<0.001f) {
			for(size_t i=0; i<bmp.getHeight(); ++i) {
				for(size_t j=0; j<bmp.getWidth(); ++j) {
					bmp[i][j].r=bmp[i][j].g=bmp[i][j].b=255;
					bmp[i][j].a=255;
				}
			}
			return;
		}
		for(size_t i=0; i<bmp.getHeight(); ++i) {
			for(size_t j=0; j<bmp.getWidth(); ++j) {
				ist::vector2 s = max_edge-min_edge;
				ist::vector2 pos = min_edge+ist::vector2(s.x/bmp.getWidth()*j, s.y/bmp.getHeight()*i);
				u_char p = u_char((getHeight(pos)-low)/(high-low)*255);
				bmp[i][j].r=bmp[i][j].g=bmp[i][j].b=p;
				bmp[i][j].a=255;
			}
		}
	}



	bool load(const std::string& fname) {
		if(fname.rfind(".bmp")!=std::string::npos
		|| fname.rfind(".tga")!=std::string::npos
		|| fname.rfind(".png")!=std::string::npos) {
			ist::Bitmap bmp;
			if(!bmp.load(fname))
				return false;
			return load(bmp);
		}
		return BGround::load(fname);
	}

	bool load(const ist::Bitmap& bmp) {
		for(int i=0; i<z_div; ++i) {
			for(int j=0; j<x_div; ++j) {
				int x = int(float(bmp.getHeight())/z_div*i);
				int z = int(float(bmp.getWidth())/x_div*j);
				float h = ist::bRGBA_to_fRGBA(bmp[x][z]).shin();
				(*this)[i][j].pos.y = h*100.0f-50.0f;
			}
		}
		setHeightTexcoord();
		setNormal();
		return true;
	}

	void scale(float xz, float y) {
		min_edge*=xz;
		max_edge*=xz;
		grid_size*=xz;
		for(size_t i=0; i<vertex.size(); ++i) {
			vertex[i].pos.x*=xz;
			vertex[i].pos.z*=xz;
			vertex[i].pos.y*=y;
		}
	}

	void draw() const {
		drawPolygon();
	}

	void drawPoint() const {
		glBegin(GL_POINTS);
		for(size_t i=0; i<vertex.size(); ++i) {
			glVertex3fv(vertex[i].pos.v);
		}
		glEnd();
	}

	void drawLine() const {
		for(size_t i=0; i<index.size()/4; ++i) {
			glBegin(GL_LINE_LOOP);
			glVertex3fv(vertex[index[i*4+0]].pos.v);
			glVertex3fv(vertex[index[i*4+1]].pos.v);
			glVertex3fv(vertex[index[i*4+2]].pos.v);
			glVertex3fv(vertex[index[i*4+3]].pos.v);
			glEnd();
		}
	}

	void drawPolygon() const {
		glBegin(GL_QUADS);
		for(size_t i=0; i<index.size(); ++i) {
			glTexCoord2fv(vertex[index[i]].tex.v);
			glNormal3fv(vertex[index[i]].nml.v);
			glVertex3fv(vertex[index[i]].pos.v);
		}
		glEnd();
	}

	void setSquareTexcoord() {
		for(int i=0; i<z_div; ++i) {
			for(int j=0; j<x_div; ++j) {
				vertex_t &v = (*this)[i][j];
				v.tex.x = 1.0f/(x_div-1)*j;
				v.tex.y = 1.0f/(z_div-1)*i;
			}
		}
	}

	void setHeightTexcoord() {
		float l=0, h=0;
		getHeightRange(l, h);
		setHeightTexcoord(l, h);
	}
	void setHeightTexcoord(float low, float high) {
		for(size_t i=0; i<vertex.size(); ++i) {
			vertex_t& v = vertex[i];
			v.tex.y = clamp(1.0f-(v.pos.y-low)/(high-low));
		}
	}


};





class Minimap
{
public:
	Minimap() : is_dragging(false), pcamera(0) {
		rect_max = ist::vector2(130, 130),
		setPos(ist::vector2(0, 0));
	}

	void resize(const ist::vector2& mmin, const ist::vector2& mmax) {
		screen_l = mmin.x/0.707f;
		screen_r = mmax.x/0.707f;
		screen_b = mmin.y/0.707f;
		screen_t = mmax.y/0.707f;
	}

	bool isMouseOver(int x, int y) {
		float ang = 0;
		ist::vector2 mouse(x, y);
		ang+=(v1-mouse).cos(v2-mouse);
		ang+=(v2-mouse).cos(v3-mouse);
		ang+=(v3-mouse).cos(v4-mouse);
		ang+=(v4-mouse).cos(v1-mouse);
		if(ang>359.0f*ist::radian)
			return true;
		return false;
	}

	ist::vector2 getMousePointerPos(int mouse_x, int mouse_y) {
		ist::vector2 rel = ist::vector2(mouse_x, mouse_y)-rect_min;
		ist::vector2 ratio = rel/(rect_max-rect_min);
		ist::vector2 pos2((screen_r-screen_l)*ratio.x+screen_l, (screen_t-screen_b)*ratio.y+screen_b);
		pos2 = ist::matrix22().rotate(-45)*pos2;
		return pos2;
	}

	void setCamera(ist::i3d::PerspectiveCamera& c) { pcamera=&c; }
	void setPos(const ist::vector2& p) {
		rect_min+=p;
		rect_max+=p;
		v1 = ist::vector2((rect_min.x+rect_max.x)/2, rect_min.y);
		v2 = ist::vector2(rect_min.x, (rect_min.y+rect_max.y)/2);
		v3 = ist::vector2((rect_min.x+rect_max.x)/2, rect_max.y);
		v4 = ist::vector2(rect_max.x, (rect_min.y+rect_max.y)/2);
	}

	void regenTexture(ist::Bitmap &bmp) {
		texture.load(bmp);
	}

	void draw() {
		glDisable(GL_DEPTH_TEST);

		glEnable(GL_TEXTURE_2D);
		texture.assign();
		{
			ist::i3d::MatrixSaver msaver(GL_MODELVIEW_MATRIX);
			ist::i3d::MatrixSaver psaver(GL_PROJECTION_MATRIX);
			ist::i3d::ScreenMatrixIV smatrix;

			glDisable(GL_DEPTH_TEST);
			glDepthMask(GL_FALSE);
			glBegin(GL_QUADS);
			glTexCoord2f(0, 0);
			glVertex2fv(v1.v);
			glTexCoord2f(0, 1);
			glVertex2fv(v2.v);
			glTexCoord2f(1, 1);
			glVertex2fv(v3.v);
			glTexCoord2f(1, 0);
			glVertex2fv(v4.v);
			glEnd();
			glDepthMask(GL_TRUE);
			glEnable(GL_DEPTH_TEST);
		}
		glDisable(GL_TEXTURE_2D);

		assert(pcamera!=0);
		ist::i3d::OrthographicCamera top_camera;
		top_camera.setScreen(screen_l, screen_r, screen_b, screen_t);
		top_camera.direction = ist::vector3(0, 0, -1.0f);
		top_camera.position = ist::vector4(0, 5000.0f, 0);
		top_camera.z_far = 10000.0f;
		top_camera.look();
		{
			ist::i3d::MatrixSaver(GL_MODELVIEW_MATRIX);
			glPointSize(5);
			glRotatef(-45, 0, 1, 0);
			glBegin(GL_POINTS);
			glVertex3fv(pcamera->target.v);
			glEnd();
			glPointSize(1);
		}
		pcamera->look();

		glEnable(GL_DEPTH_TEST);

		ist::i3d::Color().assign();
	}


	bool is_dragging;
	ist::i3d::Texture texture;
	ist::i3d::PerspectiveCamera *pcamera;
	std::vector<ist::vector4> point_pos;
	std::vector<ist::vector4> point_col;
	ist::vector2 rect_min;
	ist::vector2 rect_max;
	ist::vector2 v1, v2, v3, v4;
	float screen_l;
	float screen_r;
	float screen_b;
	float screen_t;
};






enum {
	EDIT_HEIGHT,
	EDIT_UV,
	EDIT_ATTR,

	EDIT_UP,
	EDIT_DOWN,
	EDIT_FLAT,

	EDIT_AUTOUV,
};

class Brush
{
public:
	Brush() : ground(0), type(EDIT_HEIGHT), mode(EDIT_UP), mode_uv(EDIT_AUTOUV), autouv(true), edit_u(false),
		radius(75.0f), strength(4.0f), flat(0.0f), tex_low(-100.0f), tex_high(100.0f), attr(1)
	{}

	void draw() {
		glDisable(GL_DEPTH_TEST);
		glBegin(GL_LINE_LOOP);
		int div = int(radius*0.6f)+4;
		for(int i=0; i<div; ++i) {
			float r=(360.0f/div)*i*ist::radian;
			float x=pos.x+radius*cos(r);
			float z=pos.z+radius*sin(r);
			float y=getGround().getHeight(ist::vector2(x, z));
			if(y>-5000.0f)
				glVertex3f(x, y, z);
		}
		glEnd();

		glColor3f(0.5f, 0.5f, 1.0f);
		glBegin(GL_POINTS);
		for(size_t i=0; i<marked.size(); ++i)
			glVertex3fv(marked[i]->v);
		glEnd();
		glEnable(GL_DEPTH_TEST);
		glColor3f(1.0f, 1.0f, 1.0f);
	}

	void editHeight(Ground::vertex_t& vtx) {
		float dist = (ist::vector2(vtx.pos.x, vtx.pos.z)-ist::vector2(pos.x, pos.z)).norm();
		if(dist<radius) {
			float mod = 1.0f-powf(dist/radius, 1.5f);
			if(mode==EDIT_UP)
				vtx.pos.y+=mod*strength;
			else if(mode==EDIT_DOWN)
				vtx.pos.y-=mod*strength;
			else if(mode==EDIT_FLAT)
				vtx.pos.y+=(flat-vtx.pos.y)*(strength*0.05f*mod);
			if(autouv)
				vtx.tex.y = clamp(1.0f-(vtx.pos.y-tex_low)/(tex_high-tex_low));
		}
	}

	void editUV(Ground::vertex_t& vtx) {
		float dist = (ist::vector2(vtx.pos.x, vtx.pos.z)-ist::vector2(pos.x, pos.z)).norm();
		if(dist<radius) {
			float mod = 1.0f-powf(dist/radius, 1.5f);
			float *uv;
			float *f_uv;
			if(edit_u) {
				uv = &vtx.tex.x;
				f_uv = &flat_uv.x;
			}
			else {
				uv = &vtx.tex.y;
				f_uv = &flat_uv.y;
			}

			if(mode==EDIT_UP)
				*uv-=mod*strength*0.001f;
			else if(mode==EDIT_DOWN)
				*uv+=mod*strength*0.001f;
			else if(mode==EDIT_FLAT)
				*uv+=(*f_uv-*uv)*(strength*0.05f*mod);
			*uv=clamp(*uv);
		}
	}

	void editAttr(Ground::vertex_t& vtx) {
		if((ist::vector2(vtx.pos.x, vtx.pos.z)-ist::vector2(pos.x, pos.z)).square()<radius*radius) {
			vtx.attr = attr;
			marked.push_back(&vtx.pos);
		}
	}

	void execute() {
		Ground& g=getGround();
		Ground::coord_t mn=g.getCoord(ist::vector2(pos.x-radius, pos.z-radius));
		Ground::coord_t mx=g.getCoord(ist::vector2(pos.x+radius, pos.z+radius));
		mn.x = std::min(g.getXDiv(), std::max(0, mn.x));
		mn.z = std::min(g.getXDiv(), std::max(0, mn.z));
		mx.x = std::min(g.getZDiv(), std::max(0, mx.x));
		mx.z = std::min(g.getZDiv(), std::max(0, mx.z));

		for(int i=mn.z; i<mx.z; ++i) {
			for(int j=mn.x; j<mx.x; ++j) {
				if(type==EDIT_HEIGHT)
					editHeight(g[i][j]);
				else if(type==EDIT_UV)
					editUV(g[i][j]);
				else if(type==EDIT_ATTR)
					editAttr(g[i][j]);
			}
		}
		if(type==EDIT_HEIGHT) {
			g.setNormal(mn, mx);
		}
		else if(type==EDIT_ATTR) {
			std::sort(marked.begin(), marked.end());
			std::unique(marked.begin(), marked.end());
		}
	}

	void refresh() {
		marked.clear();
		Ground& g=getGround();
		for(int i=0; i<g.getZDiv(); ++i) {
			for(int j=0; j<g.getXDiv(); ++j) {
				if(g[i][j].attr==attr)
					marked.push_back(&g[i][j].pos);
			}
		}
	}

	float getRadius() { return radius; }
	void getVRange(float& l, float& h) { l=tex_low; h=tex_high;}

	void setFlatHeight() {
		flat=pos.y;
		Ground::coord_t c=getGround().getCoord(ist::vector2(pos.x, pos.z));
		c.x = std::min(getGround().getXDiv(), std::max(0, c.x));
		c.z = std::min(getGround().getXDiv(), std::max(0, c.z));
		flat_uv=getGround()[c.z][c.x].tex;
	}
	void setGround(Ground& g) { ground=&g; }
	void setAutoUV(bool f) { autouv=f; }
	void setEditU(bool f) { edit_u=f; }
	void setPos(const ist::vector4& a) { pos=a; }
	void setRadius(float a)    { radius=a; }
	void setStrength(float a)  { strength=a; }
	void setAttr(int a)        { attr=a; refresh(); }
	void setBrushType(int t)   { type=t; }
	void setEditMode(int m)    { mode=m; }
	void setUVEditMode(int m)  { mode_uv=m; }
	void setVRange(float l, float h) { tex_low=l; tex_high=h;}

private:
	Ground& getGround() { assert(ground!=0); return *ground; }

	Ground *ground;
	std::vector<ist::vector3*> marked;
	ist::vector4 pos;
	int type;
	int mode;
	int mode_uv;
	bool autouv;
	bool edit_u;
	float radius;
	float strength;
	float flat;
	ist::vector2 flat_uv;
	float tex_low;
	float tex_high;
	int attr;
};








class GEApp
{
public:
	GEApp() {
		draw_point = false;
		draw_line = false;
		draw_polygon = true;
		lighting = true;
		g_tex_enable = true;
		brush.setGround(ground);
		minimap.setCamera(camera);

		setCameraFloating(800.0f);
		camera.fovy = 30.0f;
		camera.z_far = 3000.0f;
		camera.position = camera_floating;
		light.position.y = 30000.0f;
		light.position.z = -10000.0f;

		g_texture.setWarpS(GL_REPEAT);
		resize(ist::vector2(-500.0f, -500.0f), ist::vector2(500.0f, 500.0f), 10.0f);
	}


	void setCameraPos(const ist::vector4& npos) {
		camera.target = npos;
		camera.position = camera.target+camera_floating;
	}

	void moveCamera(const ist::vector4& npos) {
		setCameraPos(camera.target+npos);
	}

	bool loadGTexture(const std::string &fname) { return g_texture.load(fname); }


	void draw() {
		drawGround();
		drawMinimap();
	}

	void drawGround() {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		ist::i3d::Material mat;
		mat.diffuse = ist::vector4(1.0f, 1.0f, 1.0f, 1.0f);
		mat.assign();

		camera.look();
		glPointSize(2);
		if(draw_polygon) {
			if(lighting) {
				glEnable(GL_LIGHTING);
				light.enable();
			}
			if(g_tex_enable) {
				glEnable(GL_TEXTURE_2D);
				g_texture.assign();
			}
			ground.drawPolygon();
			light.disable();
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
		}

		glDisable(GL_DEPTH_TEST);
		if(draw_line) {
			ground.drawLine();
		}
		if(draw_point) {
			ground.drawPoint();
		}
		glEnable(GL_DEPTH_TEST);
	}

	void updateMinimap() {
		ist::Bitmap bmp(128, 128);
		ground.genBitmap(bmp);
		minimap.regenTexture(bmp);
	}

	void drawMinimap() {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		minimap.draw();
	}


	ist::vector4 getScreenRay(int mx, int my)
	{
		int viewport[4];
		glGetIntegerv(GL_VIEWPORT, viewport);
		ist::vector4 mouse_vec;
		float window_size = float(std::max<int>(viewport[2]-viewport[0], viewport[3]-viewport[1]));
		float cx = float(viewport[0]+viewport[2])/2.0f;
		float cy = float(viewport[1]+viewport[3])/2.0f;
		mouse_vec.x = (mx-cx)/window_size;
		mouse_vec.y = -(my-cy)/window_size;
	//	mouse_vec.z = -0.65f; // 60fovy
		mouse_vec.z = -1.4f; // 30fovy

		ist::matrix44 modelview;
		camera.look();
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview.v);
		modelview.invert();

		ist::matrix44 aim_camera;
		aim_camera.aimVector(modelview*mouse_vec-camera.position);

		return aim_camera*ist::vector4(0, 1, 0);
	}


	ist::vector4 getRayCross(int mx, int my) {
		return ground.getRayCross(camera.position, getScreenRay(mx, my));
	}

	void drawAsTexture(int w, int h) {
		int viewport[4];
		glGetIntegerv(GL_VIEWPORT, viewport);
		glViewport(0, 0, w, h);
		int width = w;
		int height = h;

		ist::i3d::OrthographicCamera top_camera;
		top_camera.setScreen(ground.getMin().x, ground.getMax().x, ground.getMin().y, ground.getMax().y);
		top_camera.direction = ist::vector3(0, 0, -1.0f);
		top_camera.position = ist::vector4(0, 5000.0f, 0);
		top_camera.z_far = 10000.0f;
		top_camera.look();

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		ist::i3d::Material mat;
		mat.diffuse = ist::vector4(1.0f, 1.0f, 1.0f, 1.0f);
		mat.assign();
		if(lighting) {
			glEnable(GL_LIGHTING);
			light.enable();
		}
		if(g_tex_enable) {
			glEnable(GL_TEXTURE_2D);
			g_texture.assign();
		}
		ground.drawPolygon();
		light.disable();
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glFlush();

		getCamera().look();
		glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
	}

	bool saveAsTexture(const std::string& fn, int width, int height) {
		std::vector<ist::bRGBA> buf(width*height);
		glReadBuffer(GL_BACK);
		glReadPixels(0,0, width,height, GL_RGBA, GL_UNSIGNED_BYTE, &buf[0][0] );

		ist::Bitmap bmp(width, height);
		for(int i=0; i<height; ++i) {
			for(int j=0; j<width; ++j) {
				bmp[i][j]=buf[width*height - (i+1)*width + j];
			}
		}
		bmp.save(fn);
		return true;
	}

	bool save(const std::string& fn) { return ground.save(fn); }

	bool load(const std::string& fn) {
		if(ground.load(fn)) {
			onLoaded();
			return true;
		}
		return false;
	}

	bool load(const ist::Bitmap& bmp) {
		if(ground.load(bmp)) {
			onLoaded();
			return true;
		}
		return false;
	}

	void onLoaded() {
		float l=0, h=0;
		ground.getHeightRange(l, h);
		brush.setVRange(l, h);
		minimap.resize(ground.getMin(), ground.getMax());
		brush.refresh();
	}

	void resize(const ist::vector2& new_min, const ist::vector2& new_max, float gs, bool preserve=false) {
		float l, h;
		brush.getVRange(l, h);
		if(preserve)
			ground.resize(new_min, new_max, gs);
		else
			ground.init(new_min, new_max, gs);
		ground.setHeightTexcoord(l, h);
		minimap.resize(ground.getMin(), ground.getMax());
		brush.refresh();
	}

	void scale(float xz, float y) {
		ground.scale(xz, y);
		ground.setNormal();
		minimap.resize(ground.getMin(), ground.getMax());

		float l, h;
		brush.getVRange(l, h);
		brush.setVRange(l*y, h*y);
	}

	void setCameraFloating(float d) { camera_floating=ist::vector4(0.5f, 1.0f, 0.5f)*d; setCameraPos(camera.target); }
	void setDrawPoint(bool b=true)  { draw_point=b; }
	void setDrawLine(bool b=true)   { draw_line=b; }
	void setDrawPolygon(bool b=true){ draw_polygon=b; }
	void setLighting(bool b=true)   { lighting=b; }
	void setCameraFloating(const ist::vector4& cf) { camera_floating=cf; }
	Brush& getBrush() { return brush; }
	Ground& getGround() { return ground; }
	Minimap& getMinimap() { return minimap; }
	ist::i3d::PerspectiveCamera& getCamera() { return camera; }
	const ist::vector4& getCameraFloating() { return camera_floating; }

	void enableGTexture(bool b) { g_tex_enable=b; }

private:
	int mode;
	bool draw_point;
	bool draw_line;
	bool draw_polygon;
	bool lighting;
	bool g_tex_enable;
	Brush brush;
	Ground ground;
	Minimap minimap;
	ist::i3d::PerspectiveCamera camera;
	ist::vector4 camera_floating;
	ist::i3d::Light light;
	ist::i3d::Texture g_texture;
};




#endif
