#ifndef IST_COLLISION_DETECTOR
#define IST_COLLISION_DETECTOR

#include <vector>
#include <algorithm>
#include "matrix.h"
#include "i3d_polygon_model.h"


namespace ist {
namespace i3d {

enum {
	CM_FALSE,
	CM_TRUE,
	CM_POINT,
	CM_LINE,
	CM_POLYGON,
	CM_POLYGON2D,
};

class ICollisionModel
{
public:
	virtual ~ICollisionModel()=0;
	virtual int getType() const=0;
	virtual void transform(const matrix44&) {}
};

class IPointCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_POINT; }
	virtual float getRadius() const=0;
	virtual const vector4& getPos() const=0;
	virtual const BoundingBox& getBoundingBox() const=0;
};

class ILineCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_LINE; }
	virtual float getRadius() const=0;
	virtual const vector4& getTail() const=0;
	virtual const vector4& getHead() const=0;
	virtual const BoundingBox& getBoundingBox() const=0;
};

class IPolygonCollisionModel : public ICollisionModel
{
public:
	typedef std::vector<vector4> vertex_cont;
	typedef std::vector<vector4> normal_cont;
	typedef std::vector<int>     index_cont;

	int getType() const { return CM_POLYGON; }
	virtual const vertex_cont& getVertex() const=0;
	virtual const normal_cont& getNormal() const=0;
	virtual const index_cont&  getIndex() const=0;
	virtual const BoundingBox& getBoundingBox() const=0;
};

class IPolygon2DCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_POLYGON2D; }
	virtual size_t getNumVertex() const=0;
	virtual const ist::vector2& getVertex(int) const=0;
	virtual float getTop() const=0;
	virtual float getBottom() const=0;
};




class NullCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_FALSE; }
};

class AlmightyCollisionModel : public ICollisionModel
{
public:
	int getType() const { return CM_TRUE; }
};


class SimplePointCollisionModel : public IPointCollisionModel
{
public:
	SimplePointCollisionModel() : radius(0) {}
	SimplePointCollisionModel(const vector4& p, float r) : pos(p), radius(r) { setBoundingBox(); }
	float getRadius() const { return radius; }
	const vector4& getPos() const { return pos; }
	const BoundingBox& getBoundingBox() const { return bb; }

	void setPos(const vector4& v) { pos=v; setBoundingBox(); }
	void setRadius(float r) { radius=r; setBoundingBox(); }
	void setBoundingBox() {
		bb.min.x = pos.x-radius;
		bb.min.y = pos.y-radius;
		bb.min.z = pos.z-radius;
		bb.max.x = pos.x+radius;
		bb.max.y = pos.y+radius;
		bb.max.z = pos.z+radius;
	}

protected:
	vector4 pos;
	float radius;
	BoundingBox bb;
};

class SimpleLineCollisionModel : public ILineCollisionModel
{
public:
	SimpleLineCollisionModel() : radius(0) {}
	float getRadius() const { return radius; }
	const vector4& getHead() const { return head; }
	const vector4& getTail() const { return tail; }
	const BoundingBox& getBoundingBox() const { return bb; }

	void setHead(const vector4& v) { head=v; setBoundingBox(); }
	void setTail(const vector4& v) { tail=v; setBoundingBox(); }
	void setRadius(float r) { radius=r; setBoundingBox(); }
	void setBoundingBox() {
		bb.min.x = std::min<float>(head.x+radius, tail.x+radius);
		bb.min.y = std::min<float>(head.y+radius, tail.y+radius);
		bb.min.z = std::min<float>(head.z+radius, tail.z+radius);
		bb.max.x = std::max<float>(head.x+radius, tail.x+radius);
		bb.max.y = std::max<float>(head.y+radius, tail.y+radius);
		bb.max.z = std::max<float>(head.z+radius, tail.z+radius);
	}

protected:
	vector4 head, tail;
	float radius;
	BoundingBox bb;
};

class SimplePolygonCollisionModel : public IPolygonCollisionModel
{
public:
	SimplePolygonCollisionModel() {}
	const vertex_cont& getVertex() const { return vertex; }
	const normal_cont& getNormal() const { return normal; }
	const index_cont&  getIndex() const  { return index; }
	const BoundingBox& getBoundingBox() const { return bb; }

	void setBoundingBox() {
		bb.min = vertex.front();
		bb.max = vertex.front();
		for(vertex_cont::iterator p=vertex.begin(); p!=vertex.end(); ++p) {
			bb.min.x = std::min<float>(bb.min.x, p->x);
			bb.min.y = std::min<float>(bb.min.y, p->y);
			bb.min.z = std::min<float>(bb.min.z, p->z);
			bb.max.x = std::max<float>(bb.max.x, p->x);
			bb.max.y = std::max<float>(bb.max.y, p->y);
			bb.max.z = std::max<float>(bb.max.z, p->z);
		}
	}

	vertex_cont vertex;
	normal_cont normal;
	index_cont  index;
	BoundingBox bb;
};

class SimplePolygon2DCollisionModel : public IPolygon2DCollisionModel
{
public:
	SimplePolygon2DCollisionModel() : top(1), bottom(-1) {}
	size_t getNumVertex() const { return vertex.size(); }
	const ist::vector2& getVertex(int i) const { return vertex[i]; }
	float getTop() const    { return top; }
	float getBottom() const { return bottom; }

	typedef std::vector<ist::vector2> vertex_cont;
	vertex_cont vertex;
	float top;
	float bottom;
};



class FrustumCollisionModel : public SimplePolygonCollisionModel
{
public:
	FrustumCollisionModel() :
		z_near(0),
		z_far(1500.0f)
	{
		vertex.resize(5);
		normal.resize(6);
		index.resize(18);

		index[0]=0; index[1]=1; index[2]=2;
		index[3]=0; index[4]=2; index[5]=3;
		index[6]=0; index[7]=3; index[8]=4;
		index[9]=0; index[10]=4; index[11]=1;
		index[12]=1; index[13]=4; index[14]=3;
		index[15]=3; index[16]=2; index[17]=1;
	}

	void setup(const vector4& _0, const vector4& _1, const vector4& _2, const vector4& _3, const vector4& _4) {

		vertex[0] = _0;
		vertex[1] = _1;
		vertex[2] = _2;
		vertex[3] = _3;
		vertex[4] = _4;
		for(size_t i=0; i<index.size()/3; ++i) {
			normal[i] = vector3().cross(
				vertex[index[i*3+1]]-vertex[index[i*3]], vertex[index[i*3+2]]-vertex[index[i*3]]).normalize();
		}

		setBoundingBox();
	}

	bool test(const vector4& center, float r) const {
		vector4 rel_pos = center-vertex[0];
		float dist = normal[4].dot(rel_pos);

		if(vertex[1]==vertex[2] || vertex[2]==vertex[3] || dist<z_near-r || dist>z_far+r)
			return false;
		if(normal[0].dot(rel_pos)>r || normal[2].dot(rel_pos)>r || normal[3].dot(rel_pos)>r || normal[1].dot(rel_pos)>r)
			return false;
		return true;
	}

	bool test(const ICollisionModel& cm) const {
		if(cm.getType()==CM_TRUE) {
			return true;
		}
		else if(cm.getType()==CM_POINT) {
			const IPointCollisionModel& pcm = static_cast<const IPointCollisionModel&>(cm);
			return test(pcm.getPos(), pcm.getRadius());
		}
		else if(cm.getType()==CM_LINE) {
			const ILineCollisionModel& lcm = static_cast<const ILineCollisionModel&>(cm);
			return test(lcm.getHead(), lcm.getRadius());
		}

		return false;
	}

	float z_near;
	float z_far;
};




class CollisionDetector
{
public:
	CollisionDetector();
	bool detect(const ICollisionModel& base, const ICollisionModel& target);

	bool result;
	float distance;
	vector4 collide_pos;
	vector4 collide_normal;

private:
	template <class T, class U> bool detect_(const T &base, const U &target);
};


} //i3d
} //ist

#endif
