#include "gl_include.h"
#include "i3d_camera.h"

namespace ist {
namespace i3d {

ViewFrustum::ViewFrustum() : z_near(1.0f), z_far(2000.0f) {
	vertex.resize(5);
	normal.resize(6);
	index.resize(18);

	index[0]=0; index[1]=1; index[2]=2;
	index[3]=0; index[4]=2; index[5]=3;
	index[6]=0; index[7]=3; index[8]=4;
	index[9]=0; index[10]=4; index[11]=1;
	index[12]=1; index[13]=4; index[14]=3;
	index[15]=3; index[16]=2; index[17]=1;
}

void ViewFrustum::setup(const vector4& pos_, const vector4& target4, const vector3& up, float z_near_, float z_far_, float fovy, float aspect) {
	float fovyx = fovy*aspect;
	front = vector3((target4-pos).v).normalize();
	pos = pos_;
	z_near = z_near_;
	z_far = z_far_;

	vector3 vx;
	vector3 vy;
	vx.cross(front, up);
	vx.normalize();
	vy.cross(vx, front);
	vy.normalize();

	matrix33 rot_t = matrix33().rotateA(vx, -90-fovy/2.0f);
	matrix33 rot_b = matrix33().rotateA(vx,  90+fovy/2.0f);
	matrix33 rot_r = matrix33().rotateA(vy,  90+fovyx/2.0f);
	matrix33 rot_l = matrix33().rotateA(vy, -90-fovyx/2.0f);
	top   = rot_t*front;
	bottom= rot_b*front;
	left  = rot_l*front;
	right = rot_r*front;
	vertex[0] = pos;
	vertex[1] = pos - (top+right)*z_far;
	vertex[2] = pos - (top+left)*z_far;
	vertex[3] = pos - (bottom+left)*z_far;
	vertex[4] = pos - (bottom+right)*z_far;
	for(size_t i=0; i<index.size()/3; ++i) {
		normal[i] = vector3().cross(
			vertex[index[i*3+1]]-vertex[index[i*3]], vertex[index[i*3+2]]-vertex[index[i*3]]).normalize();
	}

	setBoundingBox();
}

int ViewFrustum::test(const vector4& center, float r) const {
	vector4 rel_pos = center-pos;
	float dist = front.dot(rel_pos);

	if(dist<z_near-r || dist>z_far+r)
		return 0;
	float dot_top = top.dot(rel_pos);
	float dot_bottom = bottom.dot(rel_pos);
	float dot_right = right.dot(rel_pos);
	float dot_left = left.dot(rel_pos);
	if(dot_top>r || dot_bottom>r || dot_right>r || dot_left>r)
		return 0;
	else if(dot_top>-r || dot_bottom>-r || dot_right>-r || dot_left>-r)
		return 1;
	return 2;
}

int ViewFrustum::test(const ICollisionModel& cm) const {
	if(cm.getType()==CM_TRUE) {
		return 2;
	}
	else if(cm.getType()==CM_POINT) {
		const IPointCollisionModel& pcm = static_cast<const IPointCollisionModel&>(cm);
		return test(pcm.getPos(), pcm.getRadius());
	}
	else if(cm.getType()==CM_LINE) {
		const ILineCollisionModel& lcm = static_cast<const ILineCollisionModel&>(cm);
		return test(lcm.getHead(), lcm.getRadius());
	}

	return false;
}



Camera::Camera() :
	position(0.0f, 0.0f, 100.0f),
	target(0.0f, 0.0f, 0.0f),
	direction(0.0f, 1.0f, 0.0f),
	z_near(1.0f), z_far(1000.0f)
{}

Camera::~Camera() {}




OrthographicCamera::OrthographicCamera() :
	left(0.0f),
	right(100.0f),
	bottom(100.0f),
	top(0.0f)
{
	z_near = -1.0f;
}

void OrthographicCamera::look() {
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(left, right, bottom, top, z_near, z_far);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt(
		position.x, position.y, position.z,
		target.x, target.y, target.z,
		direction.x, direction.y, direction.z);
}

void OrthographicCamera::setScreen(float l, float r, float b, float t) {
	left = l;
	right = r;
	bottom = b;
	top = t;
}




PerspectiveCamera::PerspectiveCamera() :
	fovy(60.0f), aspect(1.3333333f)
{}

void PerspectiveCamera::look() {
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective(fovy, aspect, z_near, z_far);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt(
		position.x, position.y, position.z,
		target.x, target.y, target.z,
		direction.x, direction.y, direction.z);
}

void PerspectiveCamera::setViewFrustum() {
	view_frustum.setup(position, target, direction, z_near, z_far, fovy, aspect);
}

const ViewFrustum& PerspectiveCamera::getViewFrustum() const {
	return view_frustum;
}


} //gl
} //ist

