#include "gl_include.h"
#include "gl_texture.h"


namespace ist {
namespace i3d {


Texture::~Texture() { clear(); }
Texture::Texture() { setDefaultParam(); }

Texture::Texture(const std::string& filename, unsigned int cell_width, unsigned int cell_height) {
	setDefaultParam();
	load(Bitmap(filename), cell_width, cell_height);
}

Texture::Texture(const Bitmap& src, unsigned int cell_width, unsigned int cell_height) {
	setDefaultParam();
	load(src, cell_width, cell_height);
}

void Texture::setDefaultParam() {
	setDimension(GL_TEXTURE_2D);
	setType(GL_RGBA);
	setMagFilter(GL_LINEAR);	// GL_LINEAR | GL_NEAREST
	setMinFilter(GL_LINEAR);
	setWarpS(GL_CLAMP);			// GL_CLAMP | GL_REPEAT
	setWarpT(GL_CLAMP);
	setEnvMode(GL_MODULATE);	// GL_MODULATE | GL_DECAL | GL_BLEND
}

void Texture::clear() {
	if(!_tex_name.empty()) {
		glDeleteTextures(_tex_name.size(), &_tex_name[0]);
		_tex_name.clear();
	}
}

void Texture::setParam(unsigned int name) const {
	glTexParameteri(_dimension, GL_TEXTURE_WRAP_S, _warp_s);
	glTexParameteri(_dimension, GL_TEXTURE_WRAP_T, _warp_t);
	glTexParameteri(_dimension, GL_TEXTURE_MAG_FILTER, _mag_filter);
	glTexParameteri(_dimension, GL_TEXTURE_MIN_FILTER, _min_filter);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, _env_mode);
	
	float ecolor[] = {1.0f, 1.0f, 1.0f, 1.0f};
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, _env_mode);
	glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, ecolor);
}

void Texture::generate(unsigned int& name, const Bitmap& src) {
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glGenTextures(1, &name);
	glBindTexture(_dimension, name);
	setParam(name);

	unsigned int width = src.getWidth();
	unsigned int height = src.getHeight();

	if(_type==GL_RGBA) {
		glTexImage2D(_dimension, 0, 4, width, height, 0, _type, GL_UNSIGNED_BYTE, src[0][0].v);
	}
	else if(_type==GL_RGB) {
		std::vector<unsigned char> data(width*height*3);
		for(int i=0; i<height; ++i) {
			for(int j=0; j<width; ++j) {
				data[i*width*3 +j*3 +0] = src[i*width][j].r;
				data[i*width*3 +j*3 +1] = src[i*width][j].g;
				data[i*width*3 +j*3 +2] = src[i*width][j].b;
			}
		}
		glTexImage2D(_dimension, 0, 3, width, height, 0, _type, GL_UNSIGNED_BYTE, &data[0]);
	}
	else if(_type==GL_ALPHA || _type==GL_LUMINANCE) {
		std::vector<unsigned char> data(width*height);
		for(int i=0; i<height; ++i) {
			for(int j=0; j<width; ++j) {
				data[i*width +j] = src[i*width][j].a;
			}
		}
		glTexImage2D(_dimension, 0, 1, width, height, 0, _type, GL_UNSIGNED_BYTE, &data[0]);
	}
}

bool Texture::load(const std::string& filename, unsigned int cell_width, unsigned int cell_height) {
	Bitmap bmp;
	if(!bmp.load(filename))
		return false;
	return load(bmp, cell_width, cell_height);
}

bool Texture::load(const Bitmap& src, unsigned int cell_width, unsigned int cell_height) {
	clear();
	if(src.getWidth()<16)
		return false;

	if(cell_width==0 && cell_height==0) {
		_tex_name.push_back(0);
		generate(_tex_name.back(), src);
	}
	else if(cell_width>=16 && cell_height>=16){
		Bitmap cell(cell_width, cell_height);

		unsigned int repeat_x = src.getWidth()/cell_width;
		unsigned int repeat_y = src.getHeight()/cell_height;
		_tex_name.resize(repeat_x*repeat_y);

		for(unsigned int i=0; i<_tex_name.size(); i++) {
			for(unsigned int y=0; y<cell_height; y++) {
				for(unsigned int x=0; x<cell_width; x++) {
					cell[y][x] = src[(i/repeat_x)*cell_height+y][(i%repeat_x)*cell_width+x];
				}
			}
			generate(_tex_name[i], cell);
		}
	}
	else {
		return false;
	}

	return true;
}


void Texture::assign(int index) const {
	if(!_tex_name.empty() && index<_tex_name.size() ) {
		glBindTexture(_dimension, _tex_name[index]);
		setParam(_tex_name[index]);
	}
	else {
		glBindTexture(_dimension, 0);
	}
}



ITexture* ITexture::create() { return new Texture(); }

}	// i3d
}	// ist

