#ifndef I3D_POLYGON_MODEL_INCLUDE
#define I3D_POLYGON_MODEL_INCLUDE


#include <boost/smart_ptr.hpp>
#include <string>
#include <vector>
#include "ref_count.h"
#include "matrix.h"

#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif



namespace ist {
namespace i3d {


struct BoundingSphere
{
	vector4 pos;
	float radius;

	BoundingSphere() : radius(0) {}
	BoundingSphere(const vector4& p, float r) : pos(p), radius(r) {}
};


struct BoundingBox
{
	vector4 min, max;

	BoundingBox(){}
	BoundingBox(const vector4 &min_, const vector4 &max_) : min(min_), max(max_) {}

	BoundingBox getAABB(const matrix44 &trans) const;
	bool overlap(const BoundingBox& bb) const;
	bool inner_point(const BoundingSphere&) const;
	bool inner_point(const vector4& pos, float radius=0) const;
	int inner_triangle(const vector4& v1, const vector4& v2, const vector4& v3) const;
};





class IPolygonChunc
{
public:
	virtual ~IPolygonChunc() {}
	virtual void draw() const=0;

	typedef std::vector<vector4> vertex_cont;
	typedef std::vector<vector4> normal_cont;
	typedef std::vector<vector3> vnormal_cont;
	typedef std::vector<vector2> texcoord_cont;
	typedef std::vector<int>     index_cont;

	virtual const vertex_cont&   getVertex() const=0;
	virtual const normal_cont&   getNormal() const=0;
	virtual const vnormal_cont&  getVNormal() const=0;
	virtual const texcoord_cont& getTexcoord() const=0;
	virtual const index_cont&    getIndex() const=0;
	virtual const BoundingSphere& getBSphere() const=0;
	virtual const BoundingBox&    getBBox() const=0;

	virtual void addRef()=0;
	virtual void release()=0;
	virtual size_t getRefCount()=0;
};

inline void intrusive_ptr_add_ref(IPolygonChunc *p) { p->addRef(); }
inline void intrusive_ptr_release(IPolygonChunc *p) { p->release(); }


class IPolygonModel
{
public:
	typedef boost::intrusive_ptr<IPolygonChunc> chunc_ptr;

	static IPolygonModel* create();

	virtual ~IPolygonModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw() const=0;

	virtual size_t getNumChunc()=0;
	virtual chunc_ptr getChunc(int i)=0;
	virtual chunc_ptr getChunc(const std::string& chuncname)=0;
};



class IAnimatedModel
{
public:
	static IAnimatedModel* create();

	virtual ~IAnimatedModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw(const std::string& clipname, float t) const=0;

	virtual bool isExistClip(const std::string& clipname) const=0;
	virtual float beginFrame(const std::string& clipname) const=0;
	virtual float endFrame(const std::string& clipname) const=0;
	virtual const BoundingSphere& getBSphere() const=0;
	virtual const BoundingBox&    getBBox() const=0;

};


class ICachedAnimatedModel
{
public:
	static ICachedAnimatedModel* create();

	virtual ~ICachedAnimatedModel() {}
	virtual bool load(const std::string& filename)=0;
	virtual void draw(const std::string& clipname, int t) const=0;

	virtual bool isExistClip(const std::string& clipname) const=0;
	virtual int beginFrame(const std::string& clipname) const=0;
	virtual int endFrame(const std::string& clipname) const=0;
	virtual const BoundingSphere& getBSphere() const=0;
	virtual const BoundingBox&    getBBox() const=0;
};



} //i3d
} //ist

#endif
