#include <vector>
#include <stack>
#include "ist_lua.h"
using namespace std;



#ifndef _NDEBUG
	#define ERROR_CHECK( statement )\
try { statement }\
catch(const std::runtime_error& er) { throw_error(L, er.what()); }

#else
	#define ERROR_CHECK( statement ) statement
#endif




namespace ist {
namespace lua {



namespace  {
	enum {
		VECTOR2_ID = 100,
		VECTOR3_ID,
		VECTOR4_ID,
		MATRIX22_ID,
		MATRIX33_ID,
		MATRIX44_ID,
	};
	std::vector<vector2> vec2_stack;
	std::vector<vector3> vec3_stack;
	std::vector<vector4> vec4_stack;
	std::vector<matrix22> mat22_stack;
	std::vector<matrix33> mat33_stack;
	std::vector<matrix44> mat44_stack;
}

namespace math3d {
	struct stack_info
	{
		stack_info(size_t _v2, size_t _v3, size_t _v4, size_t _m22, size_t _m33, size_t _m44) :
			v2(_v2), v3(_v3), v4(_v4),
			m22(_m22), m33(_m33), m44(_m44)
		{}

		size_t v2, v3, v4;
		size_t m22, m33, m44;
	};

	std::stack<stack_info> lua_math3d_stack;


	void push_stack() {
		lua_math3d_stack.push(
			stack_info(vec2_stack.size(), vec3_stack.size(), vec4_stack.size(),
				mat22_stack.size(), mat33_stack.size(), mat44_stack.size()));
	}

	void pop_stack() {
		stack_info& info = lua_math3d_stack.top();
		vec2_stack.resize(info.v2);
		vec3_stack.resize(info.v3);
		vec4_stack.resize(info.v4);
		mat22_stack.resize(info.m22);
		mat33_stack.resize(info.m33);
		mat44_stack.resize(info.m44);
		lua_math3d_stack.pop();
	}

	int push_stack(lua_State *L) {
		push_stack();
		return 0;
	}

	int pop_stack(lua_State *L) {
		pop_stack();
		return 0;
	}


	StackSaver::StackSaver() {
		push_stack();
	}

	StackSaver::~StackSaver() {
		pop_stack();
	}
}

template<int NumSize, int IntSize> struct _math3d_identifier;

template<>
struct _math3d_identifier<8, 4> {

	union {
		double num_data;
		struct {
			unsigned int id;
			unsigned int pos;
		};
	};

	_math3d_identifier() : num_data(0) {}
	_math3d_identifier(double d) : num_data(d) {}
	_math3d_identifier(unsigned char i, unsigned int p) : id(i), pos(p) {}

	bool isNumber() {
		return (!isVector2() && !isVector3() && !isVector4() && !isMatrix22() && !isMatrix33() && !isMatrix44());
	}

	bool isVector2() { return id==VECTOR2_ID; }
	bool isVector3() { return id==VECTOR3_ID; }
	bool isVector4() { return id==VECTOR4_ID; }
	bool isMatrix22() { return id==MATRIX22_ID; }
	bool isMatrix33() { return id==MATRIX33_ID; }
	bool isMatrix44() { return id==MATRIX44_ID; }

	double& toNumber() {
		return num_data;
	}

	vector2& toVector2() {
		if(!isVector2())
			throw std::runtime_error("ist::lua::math3d_identifier::toVector2()");
		return vec2_stack[pos];
	}

	vector3& toVector3() {
		if(!isVector3())
			throw std::runtime_error("ist::lua::math3d_identifier::toVector3()");
		return vec3_stack[pos];
	}

	vector4& toVector4() {
		if(!isVector4())
			throw std::runtime_error("ist::lua::math3d_identifier::toVector4()");
		return vec4_stack[pos];
	}

	matrix22& toMatrix22() {
		if(!isMatrix22())
			throw std::runtime_error("ist::lua::math3d_identifier::toMatrix22()");
		return mat22_stack[pos];
	}

	matrix33& toMatrix33() {
		if(!isMatrix33())
			throw std::runtime_error("ist::lua::math3d_identifier::toMatrix33()");
		return mat33_stack[pos];
	}

	matrix44& toMatrix44() {
		if(!isMatrix44())
			throw std::runtime_error("ist::lua::math3d_identifier::toMatrix44()");
		return mat44_stack[pos];
	}

};

typedef _math3d_identifier<sizeof(lua_Number), sizeof(int)> math3d_identifier;



void throw_error(lua_State *L, const char *message)
{
	char buf[1024];
	lua_Debug ld_current, ld_root;

	lua_getstack(L, 0, &ld_current);
	lua_getinfo(L, "n", &ld_current);

	lua_getstack(L, 1, &ld_root);
	lua_getinfo(L, "Sl", &ld_root);

	if(message!=0)
		sprintf(buf, "lua::math3d error at %s file:%s line:%d  %s\n",
			ld_current.name, ld_root.source, ld_root.currentline, message);
	else
		sprintf(buf, "lua::math3d error at %s file:%s line:%d\n",
			ld_current.name, ld_root.source, ld_root.currentline);

	throw std::runtime_error(buf);
}


inline math3d_identifier math3d_toid(lua_State *L, int index)
{
	return math3d_identifier(lua_tonumber(L, index));
}



template<class T> inline math3d_identifier math3d_push(const T& v);

template<> inline math3d_identifier math3d_push<vector2>(const vector2& v) {
	vec2_stack.push_back(v);
	return math3d_identifier(VECTOR2_ID, vec2_stack.size()-1);
}

template<> inline math3d_identifier math3d_push<vector3>(const vector3& v) {
	vec3_stack.push_back(v);
	return math3d_identifier(VECTOR3_ID, vec3_stack.size()-1);
}

template<> inline math3d_identifier math3d_push<vector4>(const vector4& v) {
	vec4_stack.push_back(v);
	return math3d_identifier(VECTOR4_ID, vec4_stack.size()-1);
}

template<> inline math3d_identifier math3d_push<matrix22>(const matrix22& v) {
	mat22_stack.push_back(v);
	return math3d_identifier(MATRIX22_ID, mat22_stack.size()-1);
}

template<> inline math3d_identifier math3d_push<matrix33>(const matrix33& v) {
	mat33_stack.push_back(v);
	return math3d_identifier(MATRIX33_ID, mat33_stack.size()-1);
}

template<> inline math3d_identifier math3d_push<matrix44>(const matrix44& v) {
	mat44_stack.push_back(v);
	return math3d_identifier(MATRIX44_ID, mat44_stack.size()-1);
}



template<class T> inline T& math3d_to(lua_State *L, int pos);

template<> inline vector2& math3d_to<vector2>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toVector2();
}

template<> inline vector3& math3d_to<vector3>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toVector3();
}

template<> inline vector4& math3d_to<vector4>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toVector4();
}

template<> inline matrix22& math3d_to<matrix22>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toMatrix22();
}

template<> inline matrix33& math3d_to<matrix33>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toMatrix33();
}

template<> inline matrix44& math3d_to<matrix44>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).toMatrix44();
}



template<class T> inline bool math3d_is(lua_State *L, int pos);

template<> inline bool math3d_is<lua_Number>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isNumber();
}

template<> inline bool math3d_is<vector2>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isVector2();
}

template<> inline bool math3d_is<vector3>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isVector3();
}

template<> inline bool math3d_is<vector4>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isVector4();
}

template<> inline bool math3d_is<matrix22>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isMatrix22();
}

template<> inline bool math3d_is<matrix33>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isMatrix33();
}

template<> inline bool math3d_is<matrix44>(lua_State *L, int pos) {
	return math3d_identifier(lua_tonumber(L, pos)).isMatrix44();
}



template<class T> inline int math3d_assign(lua_State *L) {
	math3d_to<T>(L, 1) = math3d_to<T>(L, 2);
	return 0;
}


template<class T> inline int math3d_vec_get_x(lua_State *L) {
	lua_pushnumber( L, math3d_to<T>(L, 1).x );
	return 1;
}

template<class T> inline int math3d_vec_get_y(lua_State *L) {
	lua_pushnumber( L, math3d_to<T>(L, 1).y );
	return 1;
}

template<class T> inline int math3d_vec_get_z(lua_State *L) {
	lua_pushnumber( L, math3d_to<T>(L, 1).z );
	return 1;
}

template<class T> inline int math3d_vec_get_w(lua_State *L) {
	lua_pushnumber( L, math3d_to<T>(L, 1).w );
	return 1;
}



template<class T> inline int math3d_vec_set_x(lua_State *L) {
	math3d_to<T>(L, 1).x = lua_tonumber(L, 2); return 0;
}

template<class T> inline int math3d_vec_set_y(lua_State *L) {
	math3d_to<T>(L, 1).y = lua_tonumber(L, 2); return 0;
}

template<class T> inline int math3d_vec_set_z(lua_State *L) {
	math3d_to<T>(L, 1).z = lua_tonumber(L, 2); return 0;
}

template<class T> inline int math3d_vec_set_w(lua_State *L) {
	math3d_to<T>(L, 1).w = lua_tonumber(L, 2); return 0;
}



template<class T> inline int math3d_vec_new(lua_State *L);

template<> inline int math3d_vec_new<vector2>(lua_State *L)
{
	math3d_identifier id;

	if(lua_gettop(L)==1)
		id = math3d_push<vector2>(math3d_identifier(lua_tonumber(L, 1)).toVector2() );
	else if(lua_gettop(L)==2)
		id = math3d_push<vector2>(vector2(lua_tonumber(L, 1), lua_tonumber(L, 2)));
	else
		id = math3d_push<vector2>(vector2());

	lua_pushnumber(L, id.toNumber());
	return 1;
}

template<> inline int math3d_vec_new<vector3>(lua_State *L)
{
	math3d_identifier id;

	if(lua_gettop(L)==1)
		id = math3d_push<vector3>(math3d_identifier(lua_tonumber(L, 1)).toVector3() );
	else if(lua_gettop(L)==3)
		id = math3d_push<vector3>(vector3(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3)));
	else
		id = math3d_push<vector3>(vector3());

	lua_pushnumber(L, id.toNumber());
	return 1;
}

template<> inline int math3d_vec_new<vector4>(lua_State *L)
{
	math3d_identifier id;

	if(lua_gettop(L)==1) {
		if(isvector3(L, 1))
			id = math3d_push<vector4>(math3d_identifier(lua_tonumber(L, 1)).toVector3() );
		else if(isvector4(L, 1))
			id = math3d_push<vector4>(math3d_identifier(lua_tonumber(L, 1)).toVector4() );
	}
	else if(lua_gettop(L)==3)
		id = math3d_push<vector4>(vector4(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3)));
	else if(lua_gettop(L)==4)
		id = math3d_push<vector4>(vector4(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4)));
	else
		id = math3d_push<vector4>(vector4());

	lua_pushnumber(L, id.toNumber());
	return 1;
}



template<class T> inline int math3d_vec_split(lua_State *L);

template<> inline int math3d_vec_split<vector2>(lua_State *L)
{
	vector2 &v = math3d_to<vector2>(L, 1);
	lua_pushnumber(L, v.x);
	lua_pushnumber(L, v.y);
	return 2;
}

template<> inline int math3d_vec_split<vector3>(lua_State *L)
{
	vector3 &v = math3d_to<vector3>(L, 1);
	lua_pushnumber(L, v.x);
	lua_pushnumber(L, v.y);
	lua_pushnumber(L, v.z);
	return 3;
}

template<> inline int math3d_vec_split<vector4>(lua_State *L)
{
	vector4 &v = math3d_to<vector4>(L, 1);
	lua_pushnumber(L, v.x);
	lua_pushnumber(L, v.y);
	lua_pushnumber(L, v.z);
	lua_pushnumber(L, v.w);
	return 4;
}



template<class T> inline int math3d_vec_add(lua_State *L) {
	lua_pushnumber(L, math3d_push<T>(math3d_to<T>(L, 1)+math3d_to<T>(L, 2)).toNumber());
	return 1;
}

template<class T> inline int math3d_vec_sub(lua_State *L) {
	lua_pushnumber(L, math3d_push<T>(math3d_to<T>(L, 1)-math3d_to<T>(L, 2)).toNumber());
	return 1;
}


template<class T> inline int math3d_vec_mulnum(lua_State *L) {
	lua_pushnumber(L, math3d_push<T>(math3d_to<T>(L, 1)*lua_tonumber(L, 2)).toNumber());
	return 1;
}


template<class T> inline int math3d_vec_mul(lua_State *L) {
	if(!math3d_is<T>(L, 2))
		return math3d_vec_mulnum<T>(L);
	lua_pushnumber(L, math3d_push<T>(math3d_to<T>(L, 1)*math3d_to<T>(L, 2)).toNumber());
	return 1;
}

template<class T> inline int math3d_vec_div(lua_State *L) {
	math3d_identifier id = math3d_push<T>(math3d_to<T>(L, 1)/math3d_to<T>(L, 2));
	lua_pushnumber(L, id.toNumber());
	return 1;
}

template<class T> inline int math3d_vec_norm(lua_State *L) {
	lua_pushnumber(L, math3d_to<T>(L, 1).norm());
	return 1;
}

template<class T> inline int math3d_vec_normalize(lua_State *L) {
	math3d_to<T>(L, 1).normalize();
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}

template<class T> inline int math3d_vec_dot(lua_State *L) {
	lua_pushnumber(L, math3d_to<T>(L, 1).dot(math3d_to<T>(L, 2)));
	return 1;
}

template<class T> inline int math3d_vec_cross(lua_State *L) {
	lua_pushnumber(L, math3d_push<T>(math3d_to<T>(L, 1).cross(math3d_to<T>(L, 2), math3d_to<T>(L, 3))).toNumber());
	return 1;
}




template<class T> inline int math3d_mat_new(lua_State *L)
{
	math3d_identifier id;
	if(lua_gettop(L)==1)
		id = math3d_push<T>(math3d_to<T>(L, 1));
	else
		id = math3d_push<T>(T());

	lua_pushnumber(L, id.toNumber());
	return 1;
}



namespace lua_vector2 {
	int _new(lua_State *L) { ERROR_CHECK( return math3d_vec_new<vector2>(L); ) }

	int split(lua_State *L) { ERROR_CHECK( return math3d_vec_split<vector2>(L); ) }
	int set_x(lua_State *L) { ERROR_CHECK( return math3d_vec_set_x<vector2>(L); ) }
	int set_y(lua_State *L) { ERROR_CHECK( return math3d_vec_set_y<vector2>(L); ) }
	int get_x(lua_State *L) { ERROR_CHECK( return math3d_vec_get_x<vector2>(L); ) }
	int get_y(lua_State *L) { ERROR_CHECK( return math3d_vec_get_y<vector2>(L); ) }

	int assign(lua_State *L){ ERROR_CHECK( return math3d_assign<vector2>(L); ) }
	int add(lua_State *L)   { ERROR_CHECK( return math3d_vec_add<vector2>(L); ) }
	int sub(lua_State *L)   { ERROR_CHECK( return math3d_vec_sub<vector2>(L); ) }
	int mul(lua_State *L)   { ERROR_CHECK( return math3d_vec_mul<vector2>(L); ) }
	int mulnum(lua_State *L){ ERROR_CHECK( return math3d_vec_mulnum<vector2>(L); ) }
	int div(lua_State *L)   { ERROR_CHECK( return math3d_vec_div<vector2>(L); ) }

	int norm(lua_State *L)      { ERROR_CHECK( return math3d_vec_norm<vector2>(L); ) }
	int normalize(lua_State *L) { ERROR_CHECK( return math3d_vec_normalize<vector2>(L); ) }
	int dot(lua_State *L)       { ERROR_CHECK( return math3d_vec_dot<vector2>(L); ) }
}

namespace lua_vector3 {
	int _new(lua_State *L) { ERROR_CHECK( return math3d_vec_new<vector3>(L); ) }

	int split(lua_State *L) { ERROR_CHECK( return math3d_vec_split<vector3>(L); ) }
	int set_x(lua_State *L) { ERROR_CHECK( return math3d_vec_set_x<vector3>(L); ) }
	int set_y(lua_State *L) { ERROR_CHECK( return math3d_vec_set_y<vector3>(L); ) }
	int set_z(lua_State *L) { ERROR_CHECK( return math3d_vec_set_z<vector3>(L); ) }
	int get_x(lua_State *L) { ERROR_CHECK( return math3d_vec_get_x<vector3>(L); ) }
	int get_y(lua_State *L) { ERROR_CHECK( return math3d_vec_get_y<vector3>(L); ) }
	int get_z(lua_State *L) { ERROR_CHECK( return math3d_vec_get_z<vector3>(L); ) }

	int assign(lua_State *L){ ERROR_CHECK( return math3d_assign<vector3>(L); ) }
	int add(lua_State *L)   { ERROR_CHECK( return math3d_vec_add<vector3>(L); ) }
	int sub(lua_State *L)   { ERROR_CHECK( return math3d_vec_sub<vector3>(L); ) }
	int mul(lua_State *L)   { ERROR_CHECK( return math3d_vec_mul<vector3>(L); ) }
	int mulnum(lua_State *L){ ERROR_CHECK( return math3d_vec_mulnum<vector3>(L); ) }
	int div(lua_State *L)   { ERROR_CHECK( return math3d_vec_div<vector3>(L); ) }

	int norm(lua_State *L)      { ERROR_CHECK( return math3d_vec_norm<vector3>(L); ) }
	int normalize(lua_State *L) { ERROR_CHECK( return math3d_vec_normalize<vector3>(L); ) }
	int dot(lua_State *L)       { ERROR_CHECK( return math3d_vec_dot<vector3>(L); ) }
	int cross(lua_State *L)     { ERROR_CHECK( return math3d_vec_cross<vector3>(L); ) }
}

namespace lua_vector4 {
	int _new(lua_State *L) { ERROR_CHECK( return math3d_vec_new<vector4>(L); ) }

	int split(lua_State *L) { ERROR_CHECK( return math3d_vec_split<vector4>(L); ) }
	int set_x(lua_State *L) { ERROR_CHECK( return math3d_vec_set_x<vector4>(L); ) }
	int set_y(lua_State *L) { ERROR_CHECK( return math3d_vec_set_y<vector4>(L); ) }
	int set_z(lua_State *L) { ERROR_CHECK( return math3d_vec_set_z<vector4>(L); ) }
	int set_w(lua_State *L) { ERROR_CHECK( return math3d_vec_set_w<vector4>(L); ) }
	int get_x(lua_State *L) { ERROR_CHECK( return math3d_vec_get_x<vector4>(L); ) }
	int get_y(lua_State *L) { ERROR_CHECK( return math3d_vec_get_y<vector4>(L); ) }
	int get_z(lua_State *L) { ERROR_CHECK( return math3d_vec_get_z<vector4>(L); ) }
	int get_w(lua_State *L) { ERROR_CHECK( return math3d_vec_get_w<vector4>(L); ) }
	
	int assign(lua_State *L){ ERROR_CHECK( return math3d_assign<vector4>(L); ) }
	int add(lua_State *L)   { ERROR_CHECK( return math3d_vec_add<vector4>(L); ) }
	int sub(lua_State *L)   { ERROR_CHECK( return math3d_vec_sub<vector4>(L); ) }
	int mul(lua_State *L)   { ERROR_CHECK( return math3d_vec_mul<vector4>(L); ) }
	int mulnum(lua_State *L){ ERROR_CHECK( return math3d_vec_mulnum<vector4>(L); ) }
	int div(lua_State *L)   { ERROR_CHECK( return math3d_vec_div<vector4>(L); ) }

	int norm(lua_State *L)      { ERROR_CHECK( return math3d_vec_norm<vector4>(L); ) }
	int normalize(lua_State *L) { ERROR_CHECK( return math3d_vec_normalize<vector4>(L); ) }
	int dot(lua_State *L)       { ERROR_CHECK( return math3d_vec_dot<vector4>(L); ) }
	int cross(lua_State *L)     { ERROR_CHECK( return math3d_vec_cross<vector4>(L); ) }
}








template<class T> inline int math3d_mat_mulvec(lua_State *L);

template<>
inline int math3d_mat_mulvec<matrix22>(lua_State *L) {
	math3d_identifier id = math3d_push<vector2>(math3d_to<matrix22>(L, 1)*math3d_to<vector2>(L, 2));
	lua_pushnumber(L, id.toNumber());
	return 1;
}

template<>
inline int math3d_mat_mulvec<matrix33>(lua_State *L) {
	math3d_identifier id = math3d_push<vector3>(math3d_to<matrix33>(L, 1)*math3d_to<vector3>(L, 2));
	lua_pushnumber(L, id.toNumber());
	return 1;
}

template<>
inline int math3d_mat_mulvec<matrix44>(lua_State *L) {
	math3d_identifier id = math3d_push<vector4>(math3d_to<matrix44>(L, 1)*math3d_to<vector4>(L, 2));
	lua_pushnumber(L, id.toNumber());
	return 1;
}



template<class T> inline int math3d_mat_mul(lua_State *L) {
	if(!math3d_is<T>(L, 2))
		return math3d_mat_mulvec<T>(L);
	math3d_identifier id = math3d_push<T>(math3d_to<T>(L, 1)*math3d_to<T>(L, 2));
	lua_pushnumber(L, id.toNumber());
	return 1;
}



template<class T> inline int math3d_mat_identity(lua_State *L) {
	math3d_to<T>(L, 1).identity();
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_invert(lua_State *L) {
	math3d_to<T>(L, 1).invert();
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}



template<class T> inline int math3d_mat_scale(lua_State *L) {
	math3d_to<T>(L, 1).scale(lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}

template<>
inline int math3d_mat_scale<matrix22>(lua_State *L) {
	math3d_to<matrix22>(L, 1).scale(lua_tonumber(L, 2), lua_tonumber(L, 3));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}



template<class T> inline int math3d_mat_rotate(lua_State *L) {
	math3d_to<T>(L, 1).rotate(lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}

template<>
inline int math3d_mat_rotate<matrix22>(lua_State *L) {
	math3d_to<matrix22>(L, 1).rotate(lua_tonumber(L, 2));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_translate(lua_State *L) {
	math3d_to<T>(L, 1).translate(lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}



template<class T> inline int math3d_mat_rotate_x(lua_State *L) {
	math3d_to<T>(L, 1).rotateX(lua_tonumber(L, 2));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_rotate_y(lua_State *L) {
	math3d_to<T>(L, 1).rotateY(lua_tonumber(L, 2));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_rotate_z(lua_State *L) {
	math3d_to<T>(L, 1).rotateZ(lua_tonumber(L, 2));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_rotate_a(lua_State *L) {
	math3d_to<T>(L, 1).rotateA(math3d_to<vector3>(L, 2), lua_tonumber(L, 3));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


template<class T> inline int math3d_mat_aim_vector(lua_State *L) {
	math3d_to<T>(L, 1).aimVector(math3d_to<vector3>(L, 2));
	lua_pushnumber(L, lua_tonumber(L, 1));
	return 1;
}


namespace lua_matrix22 {
	int _new(lua_State *L)    { ERROR_CHECK( return math3d_mat_new<matrix22>(L); ) }
	int assign(lua_State *L)  { ERROR_CHECK( return math3d_assign<matrix22>(L); ) }
	int mul(lua_State *L)     { ERROR_CHECK( return math3d_mat_mul<matrix22>(L); ) }
	int mulvec(lua_State *L)  { ERROR_CHECK( return math3d_mat_mulvec<matrix22>(L); ) }
	int identity(lua_State *L){ ERROR_CHECK( return math3d_mat_identity<matrix22>(L); ) }
	int invert(lua_State *L)  { ERROR_CHECK( return math3d_mat_invert<matrix22>(L); ) }

	int scale(lua_State *L)   { ERROR_CHECK( return math3d_mat_scale<matrix22>(L); ) }
	int rotate(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate<matrix22>(L); ) }
}

namespace lua_matrix33 {
	int _new(lua_State *L)    { ERROR_CHECK( return math3d_mat_new<matrix33>(L); ) }
	int assign(lua_State *L)  { ERROR_CHECK( return math3d_assign<matrix33>(L); ) }
	int mul(lua_State *L)     { ERROR_CHECK( return math3d_mat_mul<matrix33>(L); ) }
	int mulvec(lua_State *L)  { ERROR_CHECK( return math3d_mat_mulvec<matrix33>(L); ) }
	int identity(lua_State *L){ ERROR_CHECK( return math3d_mat_identity<matrix33>(L); ) }
	int invert(lua_State *L)  { ERROR_CHECK( return math3d_mat_invert<matrix33>(L); ) }

	int scale(lua_State *L)     { ERROR_CHECK( return math3d_mat_scale<matrix33>(L); ) }
	int rotate(lua_State *L)    { ERROR_CHECK( return math3d_mat_rotate<matrix33>(L); ) }
	int rotate_x(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_x<matrix33>(L); ) }
	int rotate_y(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_y<matrix33>(L); ) }
	int rotate_z(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_z<matrix33>(L); ) }
	int rotate_a(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_a<matrix33>(L); ) }
	int aim_vector(lua_State *L){ ERROR_CHECK( return math3d_mat_aim_vector<matrix33>(L); ) }
}

namespace lua_matrix44 {
	int _new(lua_State *L)     { ERROR_CHECK( return math3d_mat_new<matrix44>(L); ) }
	int assign(lua_State *L)   { ERROR_CHECK( return math3d_assign<matrix44>(L); ) }
	int mul(lua_State *L)      { ERROR_CHECK( return math3d_mat_mul<matrix44>(L); ) }
	int mulvec(lua_State *L)   { ERROR_CHECK( return math3d_mat_mulvec<matrix44>(L); ) }
	int identity(lua_State *L) { ERROR_CHECK( return math3d_mat_identity<matrix44>(L); ) }
	int invert(lua_State *L)   { ERROR_CHECK( return math3d_mat_invert<matrix44>(L); ) }

	int translate(lua_State *L) { ERROR_CHECK( return math3d_mat_translate<matrix44>(L); ) }
	int scale(lua_State *L)     { ERROR_CHECK( return math3d_mat_scale<matrix44>(L); ) }
	int rotate(lua_State *L)    { ERROR_CHECK( return math3d_mat_rotate<matrix44>(L); ) }
	int rotate_x(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_x<matrix44>(L); ) }
	int rotate_y(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_y<matrix44>(L); ) }
	int rotate_z(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_z<matrix44>(L); ) }
	int rotate_a(lua_State *L)  { ERROR_CHECK( return math3d_mat_rotate_a<matrix44>(L); ) }
	int aim_vector(lua_State *L){ ERROR_CHECK( return math3d_mat_aim_vector<matrix44>(L); ) }
}



template<class T> inline void _pushdata(lua_State *L, const T& v) {
	lua_pushnumber(L, math3d_push<T>(v).toNumber());
}

bool isvector2(lua_State *L, int index) { return math3d_is<vector2>(L, index); }
bool isvector3(lua_State *L, int index) { return math3d_is<vector3>(L, index); }
bool isvector4(lua_State *L, int index) { return math3d_is<vector4>(L, index); }
bool ismatrix22(lua_State *L, int index) { return math3d_is<matrix22>(L, index); }
bool ismatrix33(lua_State *L, int index) { return math3d_is<matrix33>(L, index); }
bool ismatrix44(lua_State *L, int index) { return math3d_is<matrix44>(L, index); }

void pushvector2(lua_State *L, const vector2 &v)   { _pushdata<vector2>(L, v); }
void pushvector3(lua_State *L, const vector3 &v)   { _pushdata<vector3>(L, v); }
void pushvector4(lua_State *L, const vector4 &v)   { _pushdata<vector4>(L, v); }
void pushmatrix22(lua_State *L, const matrix22 &v) { _pushdata<matrix22>(L, v); }
void pushmatrix33(lua_State *L, const matrix33 &v) { _pushdata<matrix33>(L, v); }
void pushmatrix44(lua_State *L, const matrix44 &v) { _pushdata<matrix44>(L, v); }

vector2& tovector2(lua_State *L, int index)   { return math3d_to<vector2>(L, index); }
vector3& tovector3(lua_State *L, int index)   { return math3d_to<vector3>(L, index); }
vector4& tovector4(lua_State *L, int index)   { return math3d_to<vector4>(L, index); }
matrix22& tomatrix22(lua_State *L, int index) { return math3d_to<matrix22>(L, index); }
matrix33& tomatrix33(lua_State *L, int index) { return math3d_to<matrix33>(L, index); }
matrix44& tomatrix44(lua_State *L, int index) { return math3d_to<matrix44>(L, index); }


inline void RegistFunc(lua_State *L, const char *func_name, int (*func)(lua_State*)) {
	lua_pushstring(L, func_name);
	lua_pushcfunction(L, func);
	lua_settable(L, -3);
}


void math3dlibopen(lua_State *L)
{
	lua_pushstring(L, "math3d");
	lua_newtable(L);
	RegistFunc(L, "push_stack", math3d::push_stack);
	RegistFunc(L, "pop_stack", math3d::pop_stack);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector2");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector2::_new);
	RegistFunc(L, "assign", lua_vector2::assign);
	RegistFunc(L, "split", lua_vector2::split);
	RegistFunc(L, "set_x", lua_vector2::set_x);
	RegistFunc(L, "set_y", lua_vector2::set_y);
	RegistFunc(L, "get_x", lua_vector2::get_x);
	RegistFunc(L, "get_y", lua_vector2::get_y);
	RegistFunc(L, "add", lua_vector2::add);
	RegistFunc(L, "sub", lua_vector2::sub);
	RegistFunc(L, "mul", lua_vector2::mul);
	RegistFunc(L, "mulnum", lua_vector2::mulnum);
	RegistFunc(L, "div", lua_vector2::div);
	RegistFunc(L, "norm", lua_vector2::norm);
	RegistFunc(L, "normalize", lua_vector2::normalize);
	RegistFunc(L, "dot", lua_vector2::dot);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector3");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector3::_new);
	RegistFunc(L, "assign", lua_vector3::assign);
	RegistFunc(L, "split", lua_vector3::split);
	RegistFunc(L, "set_x", lua_vector3::set_x);
	RegistFunc(L, "set_y", lua_vector3::set_y);
	RegistFunc(L, "set_z", lua_vector3::set_z);
	RegistFunc(L, "get_x", lua_vector3::get_x);
	RegistFunc(L, "get_y", lua_vector3::get_y);
	RegistFunc(L, "get_z", lua_vector3::get_z);
	RegistFunc(L, "add", lua_vector3::add);
	RegistFunc(L, "sub", lua_vector3::sub);
	RegistFunc(L, "mul", lua_vector3::mul);
	RegistFunc(L, "mulnum", lua_vector3::mulnum);
	RegistFunc(L, "div", lua_vector3::div);
	RegistFunc(L, "norm", lua_vector3::norm);
	RegistFunc(L, "normalize", lua_vector3::normalize);
	RegistFunc(L, "dot", lua_vector3::dot);
	RegistFunc(L, "cross", lua_vector3::cross);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector4");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector4::_new);
	RegistFunc(L, "assign", lua_vector4::assign);
	RegistFunc(L, "split", lua_vector4::split);
	RegistFunc(L, "set_x", lua_vector4::set_x);
	RegistFunc(L, "set_y", lua_vector4::set_y);
	RegistFunc(L, "set_z", lua_vector4::set_z);
	RegistFunc(L, "set_w", lua_vector4::set_w);
	RegistFunc(L, "get_x", lua_vector4::get_x);
	RegistFunc(L, "get_y", lua_vector4::get_y);
	RegistFunc(L, "get_z", lua_vector4::get_z);
	RegistFunc(L, "get_w", lua_vector4::get_w);
	RegistFunc(L, "add", lua_vector4::add);
	RegistFunc(L, "sub", lua_vector4::sub);
	RegistFunc(L, "mul", lua_vector4::mul);
	RegistFunc(L, "mulnum", lua_vector4::mulnum);
	RegistFunc(L, "div", lua_vector4::div);
	RegistFunc(L, "norm", lua_vector4::norm);
	RegistFunc(L, "normalize", lua_vector4::normalize);
	RegistFunc(L, "dot", lua_vector4::dot);
	RegistFunc(L, "cross", lua_vector4::cross);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix22");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix22::_new);
	RegistFunc(L, "assign", lua_matrix22::assign);
	RegistFunc(L, "mul", lua_matrix22::mul);
	RegistFunc(L, "mulvec", lua_matrix22::mulvec);
	RegistFunc(L, "identity", lua_matrix22::identity);
	RegistFunc(L, "invert", lua_matrix22::invert);
	RegistFunc(L, "scale", lua_matrix22::scale);
	RegistFunc(L, "rotate", lua_matrix22::rotate);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix33");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix33::_new);
	RegistFunc(L, "assign", lua_matrix33::assign);
	RegistFunc(L, "mul", lua_matrix33::mul);
	RegistFunc(L, "mulvec", lua_matrix33::mulvec);
	RegistFunc(L, "identity", lua_matrix33::identity);
	RegistFunc(L, "invert", lua_matrix33::invert);
	RegistFunc(L, "scale", lua_matrix33::scale);
	RegistFunc(L, "rotate", lua_matrix33::rotate);
	RegistFunc(L, "rotate_x", lua_matrix33::rotate_x);
	RegistFunc(L, "rotate_y", lua_matrix33::rotate_y);
	RegistFunc(L, "rotate_z", lua_matrix33::rotate_z);
	RegistFunc(L, "rotate_a", lua_matrix33::rotate_a);
	RegistFunc(L, "aim_vector", lua_matrix33::aim_vector);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix44");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix44::_new);
	RegistFunc(L, "assign", lua_matrix44::assign);
	RegistFunc(L, "mul", lua_matrix44::mul);
	RegistFunc(L, "mulvec", lua_matrix44::mulvec);
	RegistFunc(L, "identity", lua_matrix44::identity);
	RegistFunc(L, "invert", lua_matrix44::invert);
	RegistFunc(L, "translate", lua_matrix44::translate);
	RegistFunc(L, "scale", lua_matrix44::scale);
	RegistFunc(L, "rotate", lua_matrix44::rotate);
	RegistFunc(L, "rotate_x", lua_matrix44::rotate_x);
	RegistFunc(L, "rotate_y", lua_matrix44::rotate_y);
	RegistFunc(L, "rotate_z", lua_matrix44::rotate_z);
	RegistFunc(L, "rotate_a", lua_matrix44::rotate_a);
	RegistFunc(L, "aim_vector", lua_matrix44::aim_vector);
	lua_settable(L, LUA_GLOBALSINDEX);


}






} // lua
} // ist

